import { _warn, _now, _throttle } from './tools/helpers';
import { RECORDER_PRESET } from './constants';
import ConsoleObserverClass from './observers/console';
import EventObserverClass from './observers/event';
import HttpObserverClass from './observers/http';
import DOMMutationObserverClass from './observers/mutation';
import JSErrorObserverClass from './observers/js-error';
import HistoryObserverClass from './observers/history';
import MouseObserverClass from './observers/mouse';
import NikonD7000 from './tools/NikonD7000';
export default class RecorderClass {
    constructor(options) {
        this.observers = {
            mutation: null,
            console: null,
            event: null,
            mouse: null,
            error: null,
            history: null,
            http: null
        };
        this.options = RECORDER_PRESET;
        this.trail = [];
        this.recording = false;
        this.baseTime = 0;
        this.lastSnapshot = {
            time: 0,
            index: 0
        };
        this.observeScroll = (ele) => {
            if (ele) {
                ele.addEventListener('scroll', _throttle(this.observers.event.getScrollRecord));
            }
            else {
                _warn("Element doesn't existsed!");
            }
        };
        this.pushToTrail = (record) => {
            if (!this.recording)
                return;
            const thisRecordTime = _now() - this.baseTime;
            record = { t: thisRecordTime, ...record };
            const { time: lastSnapshotTime, index: lastSnapshotIndex } = this.lastSnapshot;
            if (thisRecordTime - lastSnapshotTime >= this.options.maxTimeSpan / 2) {
                if (lastSnapshotIndex !== 0) {
                    this.trail = this.trail.slice(lastSnapshotIndex);
                }
                const snapshotRecord = this.getSnapshotRecord();
                this.trail.push(snapshotRecord);
            }
            this.trail.push(record);
        };
        this.start = () => {
            if (this.recording) {
                _warn('record already started');
                return;
            }
            this.recording = true;
            this.baseTime = _now();
            this.trail[0] = this.getSnapshotRecord();
            Object.keys(this.observers).forEach(observerName => {
                if (this.options[observerName]) {
                    ;
                    this.observers[observerName].install();
                }
            });
            window.SessionRecorder = this;
        };
        this.stop = () => {
            if (!this.recording) {
                _warn('record not started');
                return;
            }
            this.recording = false;
            this.trail.length = 0;
        };
        this.uninstallObservers = () => {
            Object.keys(this.observers).forEach(observerName => {
                ;
                this.observers[observerName].uninstall();
            });
        };
        if (options) {
            this.options = { ...this.options, ...options };
        }
        const { mutation, history, http, event, error, console: consoleOpt, mouse } = this.options;
        this.observers = {
            console: new ConsoleObserverClass(consoleOpt),
            mutation: new DOMMutationObserverClass(mutation),
            event: new EventObserverClass(event),
            mouse: new MouseObserverClass(mouse),
            http: new HttpObserverClass(http),
            error: new JSErrorObserverClass(error),
            history: new HistoryObserverClass(history)
        };
        Object.keys(this.observers).forEach((observerName) => {
            const observer = this.observers[observerName];
            observer.$on('observed', this.pushToTrail.bind(this));
        });
    }
    getSnapshotRecord() {
        this.lastSnapshot.time = _now() - (this.baseTime || _now());
        this.lastSnapshot.index = this.trail.length;
        const { clientWidth: w, clientHeight: h } = document.documentElement;
        const { x, y } = this.observers.event.getScrollPosition();
        return {
            t: this.lastSnapshot.time,
            type: 'snapshot',
            scroll: { x, y },
            resize: {
                w,
                h
            },
            snapshot: NikonD7000.takeSnapshotForPage()
        };
    }
}
