import { EventTypes } from '../models/index';
import { _throttle, _log } from '../tools/helpers';
import RecorderDocument from '../tools/NikonD7000';
import BasicObserverClass from './index';
import { OBSERVER_DEFAULT_OPTIONS } from '../constants';
const { getRecordIdByElement } = RecorderDocument;
export default class EventObserverClass extends BasicObserverClass {
    constructor(options) {
        super();
        this.name = 'EventObserverClass';
        this.listeners = [];
        this.options = OBSERVER_DEFAULT_OPTIONS.event;
        this.status = OBSERVER_DEFAULT_OPTIONS.event;
        this.addListener = ({ target, event, callback, options }, cb) => {
            target.addEventListener(event, callback, options);
            this.listeners.push({
                target,
                event,
                callback
            });
            cb && cb();
        };
        this.getScrollPosition = () => {
            const isStandardsMode = document.compatMode === 'CSS1Compat';
            const x = isStandardsMode
                ? document.documentElement.scrollLeft
                : document.body.scrollLeft;
            const y = isStandardsMode
                ? document.documentElement.scrollTop
                : document.body.scrollTop;
            return { x, y };
        };
        this.getScrollRecord = (evt) => {
            const { target } = evt || { target: document };
            const { $emit } = this;
            let record = { type: EventTypes.scroll };
            if (target === document || !target) {
                let { x, y } = this.getScrollPosition();
                record = { ...record, x, y };
                $emit('observed', record);
                return;
            }
            let targetX = target;
            const { scrollLeft: x, scrollTop: y } = targetX;
            const recorderId = getRecordIdByElement(targetX);
            record = { ...record, x, y, target: recorderId };
            $emit('observed', record);
        };
        this.getResizeRecord = () => {
            const { clientWidth: w, clientHeight: h } = document.documentElement;
            const record = { type: EventTypes.resize, w, h };
            const { $emit } = this;
            $emit('observed', record);
        };
        this.getFormChangeRecord = (evt) => {
            const { target } = evt;
            if (target.contentEditable === 'true') {
                return;
            }
            const recorderId = getRecordIdByElement(target);
            let k;
            let v;
            if (!recorderId)
                return;
            const itemsWhichKeyIsChecked = ['radio', 'checked'];
            const targetX = target;
            const { type: formType } = targetX;
            if (itemsWhichKeyIsChecked.includes(formType)) {
                k = 'checked';
                v = targetX.checked;
            }
            else {
                k = 'value';
                v = targetX.value;
            }
            const record = {
                type: EventTypes.form,
                target: recorderId,
                k,
                v
            };
            const { $emit } = this;
            $emit('observed', record);
        };
        if (options === false)
            return;
        if (typeof options === 'object') {
            this.options = { ...this.options, ...options };
        }
    }
    install() {
        const { addListener } = this;
        const { scroll, resize, form } = this.options;
        if (scroll) {
            addListener({
                target: document,
                event: 'scroll',
                callback: this.getScrollRecord,
                options: true
            });
            this.status.scroll = true;
        }
        if (resize) {
            addListener({
                target: window,
                event: 'resize',
                callback: _throttle(this.getResizeRecord)
            });
            this.status.resize = true;
        }
        if (form) {
            addListener({
                target: document,
                event: 'change',
                callback: this.getFormChangeRecord,
                options: true
            });
            addListener({
                target: document,
                event: 'input',
                callback: _throttle(this.getFormChangeRecord, 300),
                options: true
            });
            this.status.form = true;
        }
        _log('events observer ready!');
    }
    uninstall() {
        const eventName2StatusKey = {
            change: 'form'
        };
        this.listeners.forEach(({ target, event, callback }) => {
            target.removeEventListener(event, callback);
            const statusKey = eventName2StatusKey[event] || event;
            this.status[statusKey] = false;
        });
    }
}
