import { HttpRockets } from '../models/index';
import { _replace, _recover, _newuuid, _log } from '../tools/helpers';
import { isFunction } from '../tools/is';
import BasicObserverClass from './index';
import { OBSERVER_DEFAULT_OPTIONS } from '../constants';
export default class HttpObserverClass extends BasicObserverClass {
    constructor(options) {
        super();
        this.name = 'HttpObserverClass';
        this.options = OBSERVER_DEFAULT_OPTIONS.http;
        this.status = {
            beacon: false,
            fetch: false,
            xhr: false
        };
        this.xhrRecordMap = new Map();
        if (options === false)
            return;
        if (typeof options === 'object') {
            this.options = { ...this.options, ...options };
        }
    }
    isSupportBeacon() {
        return !!navigator.sendBeacon;
    }
    hijackBeacon() {
        if (!this.isSupportBeacon())
            return;
        const { $emit } = this;
        function beaconReplacement(originalBeacon) {
            return function (url, data) {
                const result = originalBeacon.call(this, url, data);
                const record = {
                    type: HttpRockets.beacon,
                    url
                };
                $emit('observed', record);
                return result;
            };
        }
        _replace(window.navigator, 'sendBeacon', beaconReplacement);
    }
    isSupportFetch() {
        return window.fetch && window.fetch.toString().includes('native');
    }
    hijackFetch() {
        if (!this.isSupportFetch())
            return;
        const { $emit } = this;
        function fetchReplacement(originalFetch) {
            return function (input, config) {
                const requestId = _newuuid();
                let _method = 'GET';
                let _url;
                if (typeof input === 'string') {
                    _url = input;
                }
                else if (input instanceof Request) {
                    const { method, url } = input;
                    _url = url;
                    if (method)
                        _method = method;
                }
                else {
                    _url = String(input);
                }
                if (config && config.method) {
                    _method = config.method;
                }
                const record = {
                    type: HttpRockets.fetch,
                    method: _method,
                    url: _url,
                    id: requestId,
                    input: [...arguments]
                };
                return (originalFetch
                    .call(window, ...arguments)
                    .then((response) => {
                    record.status = response.status;
                    record.response = response.json();
                    $emit('observed', record);
                    return response;
                })
                    .catch((error) => {
                    const { message } = error;
                    record.errmsg = message;
                    $emit('observed', record);
                    throw error;
                }));
            };
        }
        _replace(window, 'fetch', fetchReplacement);
    }
    hijackXHR() {
        const { $emit } = this;
        const self = this;
        function XHROpenReplacement(originalOpen) {
            return function (method, url) {
                const requestId = _newuuid();
                const args = [...arguments];
                let record = {
                    type: HttpRockets.xhr,
                    id: requestId,
                    url,
                    method,
                    headers: {}
                };
                this.__id__ = requestId;
                self.xhrRecordMap.set(requestId, record);
                return originalOpen.apply(this, args);
            };
        }
        function XHRSetRequestHeaderReplacement(originalSetter) {
            return function (key, value) {
                const requestId = this.__id__;
                const record = self.xhrRecordMap.get(requestId);
                if (record) {
                    record.headers[key] = value;
                }
                originalSetter.call(this, key, value);
            };
        }
        function XHRSendReplacement(originalSend) {
            return function (body) {
                const thisXHR = this;
                const { __id__: requestId, __skip_record__ } = thisXHR;
                let thisRecord = self.xhrRecordMap.get(requestId);
                if (thisRecord && !__skip_record__) {
                    thisRecord.payload = body;
                }
                function onreadystatechangeHandler() {
                    if (this.readyState === 4) {
                        if (this.__skip_record__)
                            return;
                        const record = self.xhrRecordMap.get(requestId);
                        if (record) {
                            record.status = thisXHR.status;
                            record.response = thisXHR.responseText || thisXHR.response;
                            $emit('observed', record);
                        }
                    }
                }
                if ('onreadystatechange' in thisXHR &&
                    isFunction(thisXHR.onreadystatechange)) {
                    _replace(thisXHR, 'onreadystatechange', originalStateChangeHook => {
                        return (...args) => {
                            onreadystatechangeHandler.call(thisXHR);
                            originalStateChangeHook.call(thisXHR, ...args);
                        };
                    });
                }
                else {
                    thisXHR.onreadystatechange = onreadystatechangeHandler;
                }
                try {
                    return originalSend.call(this, body);
                }
                catch (exception) {
                    const { message } = exception;
                    const record = self.xhrRecordMap.get(requestId);
                    if (record) {
                        record.errmsg = message;
                        $emit('observed', record);
                    }
                }
            };
        }
        const XHRProto = XMLHttpRequest.prototype;
        _replace(XHRProto, 'setRequestHeader', XHRSetRequestHeaderReplacement);
        _replace(XHRProto, 'open', XHROpenReplacement);
        _replace(XHRProto, 'send', XHRSendReplacement);
    }
    install() {
        const { beacon, fetch, xhr } = this.options;
        if (beacon) {
            this.hijackBeacon();
            this.status.beacon = true;
        }
        if (fetch) {
            this.hijackFetch();
            this.status.fetch = true;
        }
        if (xhr) {
            this.hijackXHR();
            this.status.xhr = true;
        }
        _log('http observer ready!');
    }
    uninstall() {
        const { beacon, fetch, xhr } = this.options;
        if (beacon) {
            _recover(window.navigator, 'sendBeacon');
            this.status.beacon = false;
        }
        if (fetch) {
            _recover(window, 'fetch');
            this.status.fetch = false;
        }
        if (xhr) {
            this.hijackBeacon();
            this.status.xhr = false;
        }
    }
}
