import { DOMMutationTypes } from '../models/index';
import { ID_KEY } from '../constants';
import NikonD7000 from '../tools/NikonD7000';
import { _log, _warn } from '../tools/helpers';
import BasicObserverClass from './index';
const { getRecordIdByElement } = NikonD7000;
export default class DOMMutationObserverClass extends BasicObserverClass {
    constructor(options) {
        super();
        this.name = 'DOMMutationObserverClass';
        this.status = {
            mutation: true
        };
        if (options === false)
            return;
    }
    process(mutationRecord) {
        try {
            const { target, attributeName } = mutationRecord;
            if (target && target.tagName === 'SCRIPT')
                return;
            switch (mutationRecord.type) {
                case 'attributes': {
                    if (attributeName === ID_KEY)
                        return;
                    return this.getAttrReocrd(mutationRecord);
                }
                case 'characterData': {
                    return this.getTextRecord(mutationRecord);
                }
                case 'childList': {
                    return this.getNodesRecord(mutationRecord);
                }
                default: {
                    return;
                }
            }
        }
        catch (error) {
            _warn(error);
        }
    }
    getAttrReocrd({ attributeName, target }) {
        let record = { attr: {} };
        record.target = getRecordIdByElement(target);
        if (record.target === undefined)
            return;
        record.type = DOMMutationTypes.attr;
        record.attr.k = attributeName;
        record.attr.v = target.getAttribute(attributeName);
        return record;
    }
    getTextRecord({ target }) {
        let record = {};
        record.type = DOMMutationTypes.text;
        record.target = getRecordIdByElement(target);
        if (!record.target) {
            const parentEle = getRecordIdByElement(target.parentElement);
            if (!parentEle) {
                return;
            }
            record.target = parentEle;
            record.html = target.parentElement.innerHTML;
        }
        else {
            record.text = target.textContent;
        }
        return record;
    }
    getNodesRecord({ target, addedNodes, removedNodes, previousSibling, nextSibling }) {
        let record = { add: [], remove: [] };
        record.target = getRecordIdByElement(target);
        if (previousSibling) {
            record.prev = getRecordIdByElement(previousSibling);
        }
        if (nextSibling) {
            record.next = getRecordIdByElement(nextSibling);
        }
        const { length: isAdd } = addedNodes;
        const { length: isRemove } = removedNodes;
        if (!isAdd && !isRemove)
            return;
        record.type = DOMMutationTypes.node;
        this.nodesFilter(addedNodes).forEach((node) => {
            let nodeData = {};
            switch (node.nodeName) {
                case '#text': {
                    nodeData.type = 'text';
                    nodeData.html = node.nodeValue;
                    if (target.childNodes.length) {
                        nodeData.index = this.getNodeIndex(node.parentElement, node);
                    }
                    break;
                }
                default: {
                    nodeData.type = 'ele';
                    const { parentElement, nodeValue } = node;
                    if (!parentElement) {
                        nodeData.html = nodeValue || node.outerHTML;
                        break;
                    }
                    nodeData.index = this.getNodeIndex(parentElement, node);
                    NikonD7000.bufferNewElement(node);
                    nodeData.html = node.outerHTML;
                    NikonD7000.unmark(node, true);
                }
            }
            if (nodeData.html === null)
                return;
            record.add.push(nodeData);
        });
        this.nodesFilter(removedNodes).forEach((node) => {
            let nodeData = {};
            switch (node.nodeName) {
                case '#text': {
                    nodeData.type = 'text';
                    const { parentElement } = node;
                    if (parentElement) {
                        nodeData.html = node.textContent;
                        nodeData.index = Array.prototype.slice
                            .call(parentElement.childNodes)
                            .indexOf(node);
                    }
                    else {
                        nodeData.textContent = node.textContent;
                    }
                    break;
                }
                default: {
                    nodeData.type = 'ele';
                    nodeData.target = getRecordIdByElement(node);
                    if (nodeData.target === undefined)
                        return;
                }
            }
            record.remove.push(nodeData);
        });
        if (!record.remove.length && !record.add.length)
            return;
        if (record.target === undefined)
            return;
        if (!record.remove.length) {
            delete record.remove;
        }
        if (!record.add.length) {
            delete record.add;
        }
        return record;
    }
    nodesFilter(nodeList) {
        return Array.prototype.slice.call(nodeList).filter(node => {
            const { nodeName, tagName } = node;
            return nodeName !== '#comment' && tagName !== 'SCRIPT';
        });
    }
    getNodeIndex(parentElement, node) {
        return Array.prototype.slice.call(parentElement.childNodes).indexOf(node);
    }
    install() {
        const mutationObserver = window.MutationObserver || window.WebKitMutationObserver;
        this.observer = new mutationObserver((records) => {
            const { $emit } = this;
            for (let record of records) {
                const DOMMutationRecord = this.process(record);
                if (DOMMutationRecord) {
                    $emit('observed', DOMMutationRecord);
                }
            }
        });
        this.observer.observe(document.documentElement, {
            attributes: true,
            childList: true,
            characterData: true,
            subtree: true
        });
        _log('mutation observer ready!');
        this.status.mutation = true;
    }
    uninstall() {
        if (this.observer) {
            this.observer.disconnect();
            this.observer = null;
        }
        this.status.mutation = false;
    }
}
